dofile("xmlgen.lua")
dofile("http.lua")
dofile("lxptree.lua")

if XMLRPC then return end

local _ = [[

XMLRPC: an XML-RPC binding for Lua.

See http://www.xmlrpc.org/ for information on XML-RPC.

XML-RPC's types don't map directly to the Lua types.  A table with the
key XMLRPC.type is a special table.  The value of the key describes
which XML-RPC type the table represents.  To construct these special
values, you should use their constructor functions rather than table
constructors.

<string>: Lua string
<i4>, <integer>: Lua number
<array>: a Lua table with a key of 1 present, or with ["n"]=0
<struct>: a Lua table (without the type key)
<boolean>: False is nil.  True must be the unique value XMLRPC.true.

<double>: Build with XMLRPC.double(123.78); access with v.double
<base64>: Build with XMLRPC.base64("BaSe64--"); access with v.base64
<dateTime.iso8601>: Build with XMLRPC.dateTime(string); access with v.dateTime

Aside from that, the examples are the documentation right now.  Have fun.

Jay Carlson
nop@nop.com
]]

_ = nil

local Public, Private = {}, {}
XMLRPC = Public

-- A unique table to designate true
Public.true = {"XMLRPC_True"}

-- A unique table to designate other special types
Public.type = {"XMLRPC_Type"}

Public.USERAGENT = "Lua XMLRPC/0.0"

Public.strict = 1

Public.registry = {}

local keep_chardata
keep_chardata = {}
keep_chardata.i4 = 1
keep_chardata.int = 1
keep_chardata.boolean = 1
keep_chardata.string = 1
keep_chardata.double = 1
keep_chardata["dateTime.iso8601"] = 1
keep_chardata.base64 = 1
keep_chardata.methodName = 1
keep_chardata.name = 1
keep_chardata.value = 1
Private.keep_chardata = keep_chardata
keep_chardata = nil

function Private.expect(l, name)
  if (l[1] ~= name) then
    error("expected " .. name .. " but found " .. l[1])
  end
end

function Public.parse_methodCall_tree(l)
  local mc = {}
  %Private.expect(l, "methodCall")
  local mn = l[2]
  %Private.expect(mn, "methodName")
  mc.name = mn[2]
  if (l[3]) then
    mc.params = %Private.parse_params(l[3])
  end
  return mc
end

function Private.parse_params(l)
  %Private.expect(l, "params")
  local params = {}
  for i = 2,getn(l) do
    local param_t = l[i]
    %Private.expect(param_t, "param")
    tinsert(params, %Private.parse_value(param_t[2]))
  end
  return params
end

Private.parse_value_type = {}
  
function Private.parse_value(l)
  %Private.expect(l, "value")
  -- if not debug1 then write(_STDERR, "----"..tostring( l).."\n") flush() end
  if type(l[2]) == "string" then return l[2] end
  local type = l[2][1]
  f = %Private.parse_value_type[type]
  if not f then error("unimplemented value type " .. type) end
  return f(l[2])
end

Private.parse_value_type.i4 = function(l)
  %Private.expect(l, "i4")
  return tonumber(l[2])
end

Private.parse_value_type.int = function(l)
  %Private.expect(l, "int")
  return tonumber(l[2])
end


Private.parse_value_type.boolean = function(l)
  %Private.expect(l, "boolean")
  if tonumber(l[2]) == 0 then
    return nil
  elseif tonumber(l[2]) == 1 then
    return %Public.true
  else
    error("invalid boolean value " .. l[2])
  end
end

Private.parse_value_type.string = function(l)
  %Private.expect(l, "string")
  return l[2] or ""
end

function Public.double(n) 
  return {[%Public.type]="double", double=n}
end

Private.parse_value_type.double = function (l)
  %Private.expect(l, "double")
  if %Public.strict then
    return %Public.double(tonumber(l[2]))
  else
    return tonumber(l[2])
  end
end

function Public.dateTime(s)
  return {[%Public.type]="dateTime", dateTime=s}
end

Private.parse_value_type["dateTime.iso8601"] = 
  function (l)
    %Private.expect(l, "dateTime.iso8601")
    if %Public.strict then
      return %Public.dateTime(l[2])
    else
      return l[2]
    end
  end

function Public.base64(s)
  return {[%Public.type]="base64", base64=s}
end

Private.parse_value_type.base64 = function(l)
  %Private.expect(l, "base64")
  if %Public.strict then
    return %Public.base64(l[2])
  else
    return l[2]
  end
end

Private.parse_value_type.array = function(l)
  %Private.expect(l, "array")
  local data = l[2]
  %Private.expect(data, "data")
  local elts = {n=0}
  for i = 2,getn(data) do
    tinsert(elts, %Private.parse_value(data[i]))
  end
  return elts
end

function Private.parse_member(l)
  %Private.expect(l, "member")
  %Private.expect(l[2], "name")
  local name = l[2][2]
  local value = %Private.parse_value(l[3])
  return name,value
end

Private.parse_value_type.struct = function (l)
  %Private.expect(l, "struct")
  local struct = {}
  for i = 2,getn(l) do
    local name,value = %Private.parse_member(l[i])
    struct[name] = value
  end
  return struct
end

function Public.parse_methodResponse_tree(l)
  %Private.expect(l, "methodResponse")
  if (l[2][1] == "params") then
    return %Private.parse_params(l[2])
  elseif (l[2][1] == "fault") then
    local val = %Private.parse_value(l[2][2])
    return nil,val
  end
end

function Private.buildtree(methodname, params)
  local base = {"methodCall", {"methodName", methodname}}
  local param_tree = {"params"}
  for i=1,getn(params) do
    local p = params[i]
    tinsert(param_tree, {"param", {"value", %Private.build_value(p)}})
  end
  if (getn(params) ~= 0) then
    tinsert(base, param_tree)
  end
  return base
end

function Private.buildresponsetree(params)
  local base = {"methodResponse"}
  local param_tree = {"params"}
  for i=1,getn(params) do
    local p = params[i]
    tinsert(param_tree, {"param", {"value", %Private.build_value(p)}})
  end
  tinsert(base, param_tree)
  return base
end

function Private.buildfaulttree(fault_struct)
  local base = {"methodResponse"}
  tinsert(base, {"fault", {"value", %Private.build_value(fault_struct)}})
  return base
end


function Private.build_value(p)
  if p == %Public.true then
    return {"boolean", "1"}
  elseif p == nil then
    return {"boolean", "0"}
  elseif type(p) == "number" then
    return {"i4", tostring(p)}
  elseif type(p) == "string" then
    return {"string", p}
  elseif type(p) == "table" then
    if p[1] or p.n == 0 then
      -- treat as array
      return %Private.build_array(p)
    elseif p[%Public.type] then
      return %Private.build_special(p)
    else
      return %Private.build_struct(p)
    end
  else
    error("unsupported transfer format")
  end
end

function Private.build_array(p)
  local values = {"data"}
  for i=1,getn(p) do
    local elt = p[i]
    local elt_tree = {"value", %Private.build_value(elt)}
    tinsert(values, elt_tree)
  end
  return {"array", values}
end

function Private.build_struct(p)
  local tree = {"struct"}
  for i,v in p do
    local elt_tree = {"member", {"name", i}, {"value", %Private.build_value(v)}}
    tinsert(tree, elt_tree)
  end
  return tree
end

function Private.build_special(p)
  local t = p[%Public.type]
  if t == "double" then
    return {"double", tostring(p.double)}
  elseif t == "dateTime" then
    return {"dateTime.iso8601", p.dateTime}
  elseif t == "base64" then
    return {"base64", p.base64}
  else
    error("unknown special type")
  end
end

function Public.callToString(methodname, params)
  local tree = %Private.buildtree(methodname, params)
  return XMLGen.treeToString(tree)
end

function Public.responseToString(params)
  local tree = %Private.buildresponsetree(params)
  return XMLGen.treeToString(tree)
end

function Public.faultToString(fault_struct)
  local tree = %Private.buildfaulttree(fault_struct)
  return XMLGen.treeToString(tree)
end

----- now, to actually parse stuff
dofile("lxptree.lua")

function Public.parseMethodCall(s)
  local p = %Private.makeParser()
  local tree = p:parse(s)
  return %Public.parse_methodCall_tree(tree)
end

function Public.parseMethodResponse(s)
  local p = %Private.makeParser()
  local tree = p:parse(s)
  return %Public.parse_methodResponse_tree(tree)
end

function Private.makeParser()
  local p = LxpTree.makeParser()
  p.filter_chardata = 1
  p.merge_chardata = 1
  p.keep_chardata = %Private.keep_chardata
  return p
end

-- Now, for really silly syntactic sugar

Private.servertag = newtag()

function Public.Server(url)
  local s = {url=url}
  settag(s, %Private.servertag)
  return s
end

Private.functiontag = newtag()

function Private.Server_gettable(table, index)
  local f = {url=rawget(table, "url"), name=index}
  settag(f, %Private.functiontag)
  return f
end

function Private.functiontag_gettable(table, index)
  local f = {url=rawget(table, "url")}
  f.name = rawget(table, "name").."."..index
  settag(f, %Private.functiontag)
  return f
end


-- Backport of a 4.1 feature to return a table as multiple return values
Private.unpack = unpack or function(arr)
                     if getn(arr)==0 then return end
                     if getn(arr)==1 then return arr[1] end
                     local firstelement = tremove(arr,1)
                     return firstelement, %Private.unpack(arr)
end

Public.USERAGENT = "Lua XMLRPC/0.0"

function Public.do_xml_rpc_call(url, host, s)
   local body, headers, status, err = HTTP.post{
      url=url, 
      headers={["content-type"]="text/xml", ["user-agent"]=%Public.USERAGENT},  
      body=s}
   -- print("body:", body)
   -- print("headers:", headers)
   -- print("status: ", status)
   -- print("error: ", err)
   -- print("-------------")
   
   if status ~= 200 then
      error("XMLRPC http error "..status..", body follows:\n"..body)
   end

   -- print(body)
   
   local res, err = XMLRPC.parseMethodResponse(body)
   if err then
      error("XMLRPC error code: "..err.faultCode.." with text "..err.faultString)
   else
      return %Private.unpack(res)
   end
end


function Private.functiontag_func(func, ...)
  local url = rawget(func, "url")
  -- print("url", rawget(func, "url"))
  -- print("name", rawget(func, "name"))
  local s = "<?xml version='1.0'?>"..%Public.callToString(rawget(func, "name"), arg)
  local _,_,host = strfind(url, "^[a-zA-Z]+:[0-9]*//([^/]+)")

  return %Public.do_xml_rpc_call(url, host, s)
end

settagmethod(Private.servertag, "gettable", Private.Server_gettable)
settagmethod(Private.functiontag, "gettable", Private.functiontag_gettable)
settagmethod(Private.functiontag, "function", Private.functiontag_func)

function Private.pack(...)
   return arg
end

function Private.findFunction(name)
  local root = %Public.registry
  while 1 do
    local match,_,first,second = strfind(name, "([^.]+)%.(.*)")
    if not match then
      return root and root[name] 
    else
      if not root[first] then return nil end
      root = root[first]
      name = second
    end
  end
end

function Private.internalDispatchString(body)
   local mc_info = %Public.parseMethodCall(body)
   local func = %Private.findFunction(mc_info.name)

   if not func then error("no such function "..mc_info.name) end
   
   local result = %Private.pack(call(func, mc_info.params))
   local result_string = XMLRPC.responseToString(result).."\n"
   return result_string
end

function Public.dispatchString(body)
   local bomb_message = {}
   file = file or _OUTPUT
   local result_string = 
      call(%Private.internalDispatchString, {body}, "x", 
           function (s) %bomb_message[1] = s end)
   if bomb_message[1] then
      result_string = XMLRPC.faultToString({faultCode=1, faultString=bomb_message[1]}).."\n"
   end

   return result_string
end
